/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*- */
/* vi: set ts=2 sw=2: */
/* conf.c
   
   GJITEN : A JAPANESE DICTIONARY FOR GNOME
  
   Copyright (C) 1999-2005 Botond Botyanszki <boti at rocketmail dot com>
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published  by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <glib/gi18n.h>

#include "conf.h"
#include "constants.h"
#include "config.h"
#include "error.h"
#include "dicutil.h"

extern gchar *kanjidicstrg[];
extern gchar *gnome_dialog_msg;
GjitenConfig conf;
GSettings *settings;
GSettings *kanji_settings;

static gchar *migrate_gconf_settings(const gchar *name) {
  gboolean needed = TRUE;
  GError *error = NULL;
  GKeyFile *kf;
  gchar **list;
  gchar *msg = NULL;
  gsize i, n;

  kf = g_key_file_new();

  g_key_file_load_from_data_dirs(kf, "gsettings-data-convert",
                                 NULL, G_KEY_FILE_NONE, NULL);
  list = g_key_file_get_string_list(kf, "State", "converted", &n, NULL);

  if (list) {
    for (i = 0; i < n; i++)
      if (! g_strcmp0(list[i], name)) {
        needed = FALSE;
        break;
      }
    g_strfreev(list);
  }
  g_key_file_free(kf);

  if (needed) {
    g_spawn_command_line_sync("gsettings-data-convert",
                              NULL, NULL, NULL, &error);
    if (error) {
      msg = g_strdup_printf(_("Could not migrate old GConf settings: %s\n"
                              "Please make sure that GConf is installed "
                              "and the gsettings-data-convert tool is in "
                              "your PATH.\nAlternatively, ignore this "
                              "message and convert your old settings "
                              "manually."), error->message);
      g_error_free(error);
    } else {
      msg = g_strdup(_("Old GConf settings were either missing or "
                       "migrated successfully."));

      /* Allow some time for the GSettings backend to record the
         changes, otherwise the default values from the new schema
         might be set.  */
      sleep(1);
    }
  }

  return msg;
}

GjitenConfig *conf_load() {
  gchar *tmpstrg;
  gchar *tmpptr;
  int i;
	GjitenDicfile *dicfile;
	GVariant *diclist;
	GVariantIter iter;
	GjitenConfig *conf;


	conf = g_new0(GjitenConfig, 1);
	conf->gconf_migration = migrate_gconf_settings("gjiten.convert");

  conf->version = g_settings_get_string(settings, "version");
	/*
  if (conf->version == NULL) { // FIXME: gconf schema
    conf->kdiccfg[KANJI] = TRUE; 
    conf->kdiccfg[RADICAL] = TRUE;
    conf->kdiccfg[STROKES] = TRUE;
    conf->kdiccfg[READING] = TRUE;
    conf->kdiccfg[ENGLISH] = TRUE;
    conf->kdiccfg[FREQ] = TRUE;   
    conf->kdiccfg[JOUYOU] = TRUE;
    conf->kdiccfg[CREF] = TRUE;
		conf->toolbar = TRUE;
    conf->menubar = TRUE;
    conf->force_ja_JP = FALSE;
    conf->force_language_c = FALSE;
		if (conf->kanjidic == NULL) conf->kanjidic = g_new0(GjitenDicfile, 1);
		conf->kanjidic->path = GJITEN_DICDIR"/kanjidic";
    conf->dictpath = GJITEN_DICDIR;
    conf->searchlimit_enabled = FALSE;
    conf->maxwordmatches = DEFMAXWORDMATCHES;
    conf->autoadjust_enabled = TRUE;
    return conf; 
  }
	*/

  conf->autoadjust_enabled = g_settings_get_boolean(settings, "autoadjust-enabled");

  conf->bigwords = g_settings_get_boolean(settings, "bigwords");
  conf->bigkanji = g_settings_get_boolean(settings, "bigkanji");
  conf->largefont = g_settings_get_string(settings, "largefont");
  conf->normalfont = g_settings_get_string(settings, "normalfont");
  conf->gdk_use_xft = g_settings_get_boolean(settings, "gdk-use-xft");
  conf->force_ja_JP = g_settings_get_boolean(settings, "force-ja-jp");
  conf->force_language_c = g_settings_get_boolean(settings, "force-language-c");
  conf->envvar_override = g_settings_get_boolean(settings, "envvar-override");

  conf->searchlimit_enabled = g_settings_get_boolean(settings, "searchlimit-enabled");
  conf->maxwordmatches = g_settings_get_int(settings, "maxwordmatches");
	if (conf->maxwordmatches < 1) {
		conf->searchlimit_enabled = FALSE;
		conf->maxwordmatches = 100;
	}

  conf->search_kata_on_hira = g_settings_get_boolean(settings, "search-kata-on-hira");
  conf->search_hira_on_kata = g_settings_get_boolean(settings, "search-hira-on-kata");
  conf->verb_deinflection = g_settings_get_boolean(settings, "deinflection-enabled");

  if (conf->kanjidic == NULL) conf->kanjidic = g_new0(GjitenDicfile, 1);
  conf->kanjidic->path = g_settings_get_string(kanji_settings, "kanjidicfile");
	if ((conf->kanjidic->path == NULL) || (strlen(conf->kanjidic->path)) == 0) {
		conf->kanjidic->path = GJITEN_DICDIR"/kanjidic";
	}
  conf->unicode_radicals = g_settings_get_boolean(kanji_settings, "unicode-radicals");
	

  conf->kanjipad = g_settings_get_string(settings, "kanjipad");
	if (conf->kanjipad == NULL) conf->kanjipad = "/usr/bin/kanjipad";

	if (conf->dicfile_list != NULL) {
		dicutil_unload_dic();
		dicfile_list_free(conf->dicfile_list);
		conf->dicfile_list = NULL;
	}

	diclist = g_settings_get_value(settings, "dictionary-list");
	g_variant_iter_init(&iter, diclist);
	while (g_variant_iter_next(&iter, "(&s&s)", &tmpstrg, &tmpptr)) {
		if (tmpstrg != NULL) {
			dicfile = g_new0(GjitenDicfile, 1);
			dicfile->path = g_strdup(tmpstrg);
			dicfile->name = g_strdup(tmpptr);
			//				printf("%s\n%s\n", tmpstrg, tmpptr);
			conf->dicfile_list = g_slist_append(conf->dicfile_list, dicfile);
		}
	}
	g_variant_unref(diclist);

	if (conf->dicfile_list != NULL) conf->selected_dic = conf->dicfile_list->data;

  //Load kanji info settings
  for (i = 0; i < KCFGNUM; i++) { 
    conf->kdiccfg[i] = g_settings_get_boolean(kanji_settings, kanjidicstrg[i]);
      /* printf("%s : %d\n",kanjidicstrg[i], conf->kdiccfg[i]); */
  }
 
  //Load gjiten search history
  conf->history = g_settings_get_strv(settings, "history");

  return conf;
}

void conf_save(GjitenConfig *conf) {
  int i;
	GSList *diclist;
	GjitenDicfile *dicfile;
	GVariantBuilder builder;

	/* Apply all changes at once.  */
	g_settings_delay(settings);
	g_settings_delay(kanji_settings);

  g_settings_set_string(settings, "version", VERSION);
  //Save kanjidic display options
  for (i = 0; i < KCFGNUM; i++) { 
    g_settings_set_boolean(kanji_settings, kanjidicstrg[i], conf->kdiccfg[i]);
  }
  
  //gconf_client_set_bool(gconf_client, "/apps/gjiten/general/tooltips", conf->tooltips, NULL);
  g_settings_set_string(kanji_settings, "kanjidicfile", conf->kanjidic->path);
  g_settings_set_boolean(kanji_settings, "unicode-radicals", conf->unicode_radicals);

	if (conf->kanjipad == NULL) conf->kanjipad = "";
  g_settings_set_string(settings, "kanjipad", conf->kanjipad);

  g_settings_set_boolean(settings, "bigwords", conf->bigwords);
  g_settings_set_boolean(settings, "bigkanji", conf->bigkanji);
  g_settings_set_string(settings, "largefont", conf->largefont == NULL ? "" : conf->largefont);
  g_settings_set_string(settings, "normalfont", conf->normalfont == NULL ? "" : conf->normalfont);
  g_settings_set_boolean(settings, "gdk-use-xft", conf->gdk_use_xft);
  g_settings_set_boolean(settings, "force-ja-jp", conf->force_ja_JP);
  g_settings_set_boolean(settings, "force-language-c", conf->force_language_c);
  g_settings_set_boolean(settings, "envvar-override", conf->envvar_override);

  g_settings_set_boolean(settings, "search-kata-on-hira", conf->search_kata_on_hira);
  g_settings_set_boolean(settings, "search-hira-on-kata", conf->search_hira_on_kata);
  g_settings_set_boolean(settings, "deinflection-enabled", conf->verb_deinflection);

  //Save dicfiles [path and name seperated with linebreak]
	diclist = conf->dicfile_list;
	g_variant_builder_init(&builder, G_VARIANT_TYPE("a(ss)"));
	while (diclist != NULL) {
		if (diclist->data == NULL) break;
		dicfile = diclist->data;
		g_variant_builder_add(&builder, "(ss)", dicfile->path, dicfile->name);
		diclist = g_slist_next(diclist);
	}
	g_settings_set_value(settings, "dictionary-list", g_variant_builder_end(&builder));

	g_settings_apply(settings);
	g_settings_apply(kanji_settings);
}

void conf_save_history(GList *history, GjitenConfig *conf) {
  gchar **vector;
  guint size;
  int i;
  if (history != NULL) {
    size = g_list_length(history);
    vector = g_new0(gchar *, size + 1);
    for (i = 0; i <= 50 && i <= size; i++) {
      vector[i] = history->data;
      history = g_list_next(history);
      if (history == NULL) break;
    } 
    g_settings_set_strv(settings, "history", (const gchar **)vector);
    g_strfreev(vector);
  }
}

void conf_save_options(GjitenConfig *conf) {
    g_settings_delay(settings);
    g_settings_set_boolean(settings, "autoadjust-enabled", conf->autoadjust_enabled);
    g_settings_set_boolean(settings, "searchlimit-enabled", conf->searchlimit_enabled);
    g_settings_set_int(settings, "maxwordmatches", conf->maxwordmatches);
    g_settings_apply(settings);
}

gboolean conf_init_handler() {
  if (settings == NULL) {
		settings = g_settings_new("net.sf.gjiten");
		kanji_settings = g_settings_get_child(settings, "kanjidic");
	}

	return TRUE;

}

void conf_close_handler() {
	if (settings != NULL) {
		GJITEN_DEBUG("calling g_object_unref(settings) [%d]\n", (int) settings);
		g_object_unref(settings);
		g_object_unref(kanji_settings);
		settings = NULL;
		kanji_settings = NULL;
	}
}

void conf_display_migration_info(GjitenConfig *conf, GtkWindow *parent) {
	GtkWidget *dlg;

	dlg = gtk_message_dialog_new(parent, GTK_DIALOG_DESTROY_WITH_PARENT,
	                             GTK_MESSAGE_INFO, GTK_BUTTONS_CLOSE,
	                             "%s", conf->gconf_migration);
	g_free(conf->gconf_migration);
	conf->gconf_migration = NULL;
	gtk_dialog_run(GTK_DIALOG(dlg));
	gtk_widget_destroy(dlg);
}
